
import datetime as dt
from abc import abstractmethod
from logging import getLogger

from ...lib.date import local_date_to_time
from ...lib.date import to_date
from ...squeal import ActivityGroup, ActivityJournal
from ...squeal.tables.pipeline import Pipeline, PipelineType
from ...stoats.pipeline import BasePipeline

log = getLogger(__name__)


def display_pipeline(session, factory, date, diary, schedule=None):
    '''
    schedule only sent for summary views.
    '''
    date = to_date(date)   # why is this needed?
    for pipeline in Pipeline.all(session, PipelineType.DIARY):
        log.info('Building %s (%s, %s)' % (pipeline.cls, pipeline.args, pipeline.kargs))
        yield from pipeline.cls(*pipeline.args, diary=diary, **pipeline.kargs). \
            display(session, factory, date, schedule=schedule)


class Displayer(BasePipeline):

    def display(self, s, f, date, schedule=None):
        if schedule:
            yield from self._display_schedule(s, f, date, schedule=schedule)
        else:
            yield from self._display_date(s, f, date)

    @abstractmethod
    def _display_schedule(self, s, f, date, schedule=None):
        raise NotImplementedError()

    @abstractmethod
    def _display_date(self, s, f, date):
        raise NotImplementedError()


class JournalDiary(Displayer):

    def _display_date(self, s, f, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity_group in s.query(ActivityGroup).order_by(ActivityGroup.sort).all():
            for ajournal in s.query(ActivityJournal). \
                    filter(ActivityJournal.finish >= start,
                           ActivityJournal.start < finish,
                           ActivityJournal.activity_group == activity_group). \
                    order_by(ActivityJournal.start).all():
                yield from self._journal_date(s, f, ajournal, date)

    @abstractmethod
    def _journal_date(self, s, f, ajournal, date):
        raise NotImplementedError()




import asyncio
from logging import getLogger
from os import makedirs, getpid
from os.path import join, expanduser
from threading import Thread, Event
from time import sleep

from notebook.notebookapp import NotebookApp
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from ..squeal import SystemConstant, SystemProcess
from ..commands.args import ROOT

log = getLogger(__name__)

STARTUP_SLEEP = 3

__RUNNING = False
__ARGS = None
__DIR = None
__URL = None


def set_server_args(args):
    global __ARGS
    __ARGS = args


def start_local():
    global __RUNNING, __URL, __DIR
    if not __RUNNING:
        __DIR, __URL = _start()
        __RUNNING = True


def stop_local():
    _stop()
    __DIR, __URL = None, None


def start_remote(s, command_callback, max_retries=5, sleep_secs=1):
    global __URL, __DIR
    s.commit()  # clear pending writes
    if SystemProcess.exists_any(s, JupyterServer):
        log.debug('Jupyter already running')
    else:
        command_callback()
        retries = 0
        while not SystemProcess.exists_any(s, JupyterServer):
            retries += 1
            if retries > max_retries:
                raise Exception('Jupyter server did not start')
            sleep(sleep_secs)
    __URL = SystemConstant.get(s, SystemConstant.JUPYTER_URL)
    __DIR = SystemConstant.get(s, SystemConstant.JUPYTER_DIR)


def start_service(s):
    dir, url = _start()
    s.add(SystemProcess(pid=getpid(), owner=JupyterServer))
    s.commit()
    SystemConstant.set(s, SystemConstant.JUPYTER_URL, url, force=True)
    SystemConstant.set(s, SystemConstant.JUPYTER_DIR, dir, force=True)


def stop_service(s):
    stop_local()
    SystemProcess.delete_all(s, JupyterServer)
    SystemConstant.delete(s, SystemConstant.JUPYTER_URL)
    SystemConstant.delete(s, SystemConstant.JUPYTER_DIR)


def connection_url():
    if __URL:
        return __URL
    else:
        raise Exception('Jupyter server not started')


def notebook_dir():
    if __DIR:
        return __DIR
    else:
        raise Exception('Jupyter server not started')


def _start():
    global __RUNNING, __ARGS
    dir = expanduser(join(__ARGS[ROOT], 'notebooks'))
    if not __RUNNING:
        log.debug(f'Creating {dir}')
        makedirs(dir, exist_ok=True)

        started = Event()

        def start():
            log.info('Starting Jupyter server in separate thread')
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
            JupyterServer.launch_instance(['--notebook-dir', dir], started=started)

        t = Thread(target=start)
        t.daemon = True
        t.start()
        started.wait()  # set in JupyterServer.start() which is as late as we can get in startup
        log.debug('Separate thread started')
        sleep(STARTUP_SLEEP)  # annoying, but we seem to need extra time or web.open() doesn't work
        __RUNNING = True

    url = JupyterServer._instance.connection_url
    return dir, url


def _stop():
    global __RUNNING
    if __RUNNING:
        try:
            log.info('Stopping Jupyter server')
            JupyterServer._instance.stop()
            log.debug('Jupyter server stopped')
        except Exception as e:
            log.warning(f'Error stopping Jupyter: {e}')
        finally:
            __RUNNING = False
    else:
        log.debug('Jupyter server not running')


class JupyterServer(NotebookApp):

    def __init__(self, started=None, **kwargs):
        self._started = started
        super().__init__(**kwargs)

    @property
    def log_level(self):
        # this silences jupyter's logging
        # the existing mechanism only lets you set a value of 50, which means that "critical" messages
        # are still logged, and the decidedly non-critical usage instructions are printed.
        return 60

    def init_signal(self):
        log.debug('Skipping signal init')

    def start(self):
        self._started.set()
        super().start()




from abc import ABC, abstractmethod
from collections import defaultdict


class DBSCAN(ABC):

    def __init__(self, epsilon, minpts):
        self.__epsilon = epsilon
        self.__minpts = minpts

    def run(self, candidates):
        count, label = self.scan(candidates)
        return sorted(self.prune(count, label), reverse=True,
                      key=lambda g: len(g) + (min(g) / max(g)))  # try to give repeatability

    def prune(self, count, label):
        index = defaultdict(list)
        for key in label.keys():
            if label[key]:
                index[label[key]].append(key)
        for i in range(count):
            if len(index[i+1]) < self.__minpts:
                del index[i+1]
        return list(index.values())

    def scan(self, candidates):
        label = defaultdict(lambda: None)
        count = 0
        for candidate in candidates:
            if label[candidate] is None:
                neighbours = list(self.neighbourhood(candidate, self.__epsilon))
                if len(neighbours) >= self.__minpts:
                    count += 1
                    label[candidate] = count
                    self.grow(count, neighbours, label)
                else:
                    label[candidate] = 0
        return count, label

    def grow(self, count, stack, label):
        n = 0
        while stack:
            candidate = stack.pop()
            neighbours = list(self.neighbourhood(candidate, self.__epsilon))
            if len(neighbours) >= self.__minpts:
                for neighbour in neighbours:
                    if not label[neighbour]:
                        if label[neighbour] is None:
                            stack.append(neighbour)
                        label[neighbour] = count
                        n += 1
            elif label[candidate] is None:
                label[candidate] = 0  # we know this is a leaf so save some time

    @abstractmethod
    def neighbourhood(self, candidate, epsilon):
        raise NotImplementedError()




from ..commands.args import PATH, FORCE, FAST, parse_pairs, CONSTANT, KARG, WORKER
from ..squeal import PipelineType
from ..stoats.pipeline import run_pipeline


def activities(args, db):
    '''
## activities

    > ch2 activities PATH [PATH ...]

Read activities data from FIT files.

### Examples

    > ch2 activities -D 'Bike=Cotic Soul' ~/fit/2018-01-01.fit

will load the give file and create an entry for the `Bike` statistic with value `Cotic Soul`
(this particular variable is used to identify bike-specific parameters for power calculation,
but arbitrary names and values can be used).

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    run_pipeline(db, PipelineType.ACTIVITY, paths=args[PATH], force=args[FORCE],
                 worker=args[WORKER] is not None, id=args[WORKER],
                 constants=args[CONSTANT], **parse_pairs(args[KARG]))
    if not args[FAST] and args[WORKER] is None:
        # don't force this - it auto-detects need
        run_pipeline(db, PipelineType.STATISTIC)




from .default import default
from ..config.database import Counter, add_topic, add_child_topic, add_topic_field
from ..squeal import StatisticJournalType
from ..uweird.fields.topic import Text


def acooke(db):

    default(db)

    with db.session_context() as s:

        c = Counter(100)

        injuries = add_topic(s, 'Inuries', c)

        ms = add_child_topic(s, injuries, 'Multiple Sclerosis', c)
        add_topic_field(s, ms, 'Notes', c, StatisticJournalType.TEXT,
                        display_cls=Text)
        add_child_topic(s, ms, 'Betaferon', c,
                        schedule='2018-08-07/2d[1]')  # reminder to take meds on alternate days

        leg = add_child_topic(s, injuries, 'Broken Femur LHS', c,
                              schedule='2018-03-11-')
        add_topic_field(s, leg, 'Notes', c, StatisticJournalType.TEXT,
                        display_cls=Text)
        add_child_topic(s, leg, 'Learn to manage tendon pain', c)  # aims added as child topics
        add_child_topic(s, leg, 'Maintain fitness', c)
        add_child_topic(s, leg, 'Visit UK', c,
                        schedule='-2018-08-11')



from logging import getLogger

from sqlalchemy import not_

from . import ActivityJournalCalculatorMixin, WaypointCalculatorMixin, MultiProcCalculator
from ..names import SEGMENT_TIME, LATITUDE, LONGITUDE, DISTANCE, S, summaries, MIN, MSR, CNT, HEART_RATE, \
    SEGMENT_HEART_RATE, BPM, MAX
from ..waypoint import filter_none
from ...squeal import ActivityJournal, SegmentJournal, StatisticJournalFloat, Timestamp

log = getLogger(__name__)


class SegmentCalculator(ActivityJournalCalculatorMixin, WaypointCalculatorMixin, MultiProcCalculator):

    def _names(self):
        return {LATITUDE: 'lat',
                LONGITUDE: 'lon',
                DISTANCE: 'distance',
                HEART_RATE: 'hr'}

    def _missing(self, s):
        # extends superclass with restriction on activities that have a segment
        existing_ids = s.query(Timestamp.key). \
            filter(Timestamp.owner == self.owner_out)
        segment_ids = s.query(SegmentJournal.activity_journal_id)
        q = s.query(ActivityJournal.start). \
            filter(not_(ActivityJournal.id.in_(existing_ids)),
                   ActivityJournal.id.in_(segment_ids))
        return [row[0] for row in self._delimit_query(q)]

    def _calculate_results(self, s, ajournal, waypoints, loader):
        for sjournal in s.query(SegmentJournal). \
                filter(SegmentJournal.activity_journal == ajournal).all():
            loader.add(SEGMENT_TIME, S, summaries(MIN, CNT, MSR), sjournal.segment, sjournal,
                       (sjournal.finish - sjournal.start).total_seconds(), sjournal.start, StatisticJournalFloat)
            waypoints = [w for w in filter_none(self._names().values(), waypoints)
                         if sjournal.start <= w.time <= sjournal.finish]
            # weight by time gap so we don't bias towards more sampled times
            gaps = [(w1.time - w0.time, 0.5 * (w0.hr + w1.hr))
                    for w0, w1 in zip(waypoints, waypoints[1:])]
            if gaps:
                weighted = sum(dt.total_seconds() * hr for dt, hr in gaps)
                average = weighted / sum(dt.total_seconds() for dt, _ in gaps)
                loader.add(SEGMENT_HEART_RATE, BPM, summaries(MAX, CNT, MSR), sjournal.segment, sjournal,
                           average, sjournal.start, StatisticJournalFloat)
            else:
                log.warning('No Heart Rate data')




from logging import getLogger

import numpy as np
import pandas as pd

from .frame import linear_resample, median_dt, present, linear_resample_time
from ..stoats.names import HEART_RATE, MAX_MED_HR_M, POWER_ESTIMATE, ACTIVE_DISTANCE, ACTIVE_TIME, \
    ACTIVE_SPEED, TIMESPAN_ID, TIME, DISTANCE, MIN_KM_TIME, MED_KM_TIME, PERCENT_IN_Z, TIME_IN_Z, HR_ZONE, MAX_MEAN_PE_M

log = getLogger(__name__)

MAX_MINUTES = (5, 10, 30, 60, 90, 120, 180)


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)


def active_stats(df):
    stats = {ACTIVE_DISTANCE: 0, ACTIVE_TIME: 0, ACTIVE_SPEED: 0}
    for timespan in df[TIMESPAN_ID].dropna().unique():
        slice = df.loc[df[TIMESPAN_ID] == timespan]
        stats[ACTIVE_DISTANCE] += slice[DISTANCE].max() - slice[DISTANCE].min()
        stats[ACTIVE_TIME] += (slice.index.max() - slice.index.min()).total_seconds()
    stats[ACTIVE_SPEED] = 3.6 * stats[ACTIVE_DISTANCE] / stats[ACTIVE_TIME]
    return stats


def times_for_distance(df, km=None, delta=10):
    stats, km = {}, km or round_km()
    t4d = pd.DataFrame({TIME: df.index}, index=df[DISTANCE])
    t4d = t4d[~t4d.index.duplicated(keep='last')]
    t4d[TIME] = (t4d[TIME] - t4d[TIME].iloc[0]).astype(np.int64) / 1e9
    lt4d = linear_resample(t4d, d=delta)
    for target in km:
        n = target * 1000 / delta
        dlt4d = lt4d.diff(periods=n).dropna()
        if present(dlt4d, TIME):
            stats[MIN_KM_TIME % target] = dlt4d[TIME].min()
            stats[MED_KM_TIME % target] = dlt4d[TIME].median()
    return stats


def hrz_stats(df, zones=None):
    stats, zones = {}, zones or range(7)
    if present(df, HR_ZONE):
        ldf = linear_resample_time(df, with_timespan=True)
        hrz = pd.cut(ldf[HR_ZONE], bins=zones).value_counts()
        dt, total = median_dt(ldf), hrz.sum()
        for interval, count in hrz.iteritems():
            zone = interval.right
            stats[PERCENT_IN_Z % zone] = 100 * count / total
            stats[TIME_IN_Z % zone] = dt * count
    return stats


def max_mean_stats(df, params=((POWER_ESTIMATE, MAX_MEAN_PE_M),), mins=None, delta=10, zero=0):
    stats, mins = {}, mins or MAX_MINUTES
    ldf = linear_resample_time(df, dt=delta, with_timespan=True, keep_nan=True)
    for name, template in params:
        ldf.loc[ldf[TIMESPAN_ID].isnull(), [name]] = zero
        cumsum = ldf[name].cumsum()
        for target in mins:
            n = (target * 60) // delta
            diff = cumsum.diff(periods=n).dropna()
            if present(diff, name):
                stats[template % target] = diff.max() / n
    return stats


def max_med_stats(df, params=((HEART_RATE, MAX_MED_HR_M),), mins=None, delta=10, gap=0.01):
    stats, mins = {}, mins or MAX_MINUTES
    ldf_all = linear_resample_time(df, dt=delta, with_timespan=False, add_time=False)
    ldf_all.interpolate('nearest')
    ldf_tstamp = ldf_all.loc[ldf_all[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())].copy()
    ldf_tstamp['gap'] = ldf_tstamp.index.astype(np.int64) / 1e9
    ldf_tstamp['gap'] = ldf_tstamp['gap'].diff()
    log.debug(f'Largest gap is {ldf_tstamp["gap"].max()}s')
    for target in mins:
        n = target * 60 // delta
        log.debug(f'Target {target}m is {n} samples (delta {delta}s)')
        splits, remain = [], ldf_all.copy()
        max_gap = max(gap * target * 60, 1.5 * delta)
        for after in ldf_tstamp.index[ldf_tstamp['gap'] > max_gap].tolist():
            before = ldf_tstamp.index[ldf_tstamp.index.get_loc(after)-1]
            splits.append(remain.loc[:before])
            remain = remain.loc[after:]
        splits.append(remain)
        log.debug(f'Split data into {len(splits)} sections for {target}m with max gap of {max_gap}s')
        for name, template in params:
            stat_name = template % target
            for split in splits:
                split['med'] = split[name].rolling(n).median()
                if present(split, 'med'):
                    max_med = split['med'].dropna().max()
                    if stat_name in stats:
                        stats[stat_name] = max(stats[stat_name], max_med)
                    else:
                        stats[stat_name] = max_med
    return stats


# if __name__ == '__main__':
#     from ch2.data import *
#     date = '2017-02-07 07:18:50'
#     s = session('-v5')
#     df = activity_statistics(s, HEART_RATE, local_time=date, activity_group_name='Bike', with_timespan=True)
#     print(max_med(df))




import datetime as dt
from logging import getLogger

from sqlalchemy.sql import func

from . import MultiProcCalculator, CompositeCalculatorMixin
from ..names import STEPS, REST_HR, HEART_RATE, DAILY_STEPS, BPM, STEPS_UNITS, summaries, SUM, AVG, CNT, MIN, MAX, MSR
from ...lib import local_date_to_time
from ...squeal import MonitorJournal, StatisticJournalInteger, StatisticName

log = getLogger(__name__)
QUARTER_DAY = 6 * 60 * 60


class MonitorCalculator(CompositeCalculatorMixin, MultiProcCalculator):

    def _missing(self, s):
        missing = super()._missing(s)
        if missing:
            # we may need to back up one day since the last data could be incomplete
            start = missing[0]
            if s.query(MonitorJournal).filter(MonitorJournal.start < local_date_to_time(start)).limit(1).one_or_none():
                log.debug('Back-up one day')
                start = start - dt.timedelta(days=1)
                missing = [start] + missing
        return missing

    def _unused_sources_given_used(self, s, used_sources):
        return self._unused_sources_given_used_and_class(s, used_sources, MonitorJournal)

    def _read_data(self, s, start, finish):
        midpt = start + dt.timedelta(hours=12)
        m0 = s.query(func.avg(func.abs(StatisticJournalInteger.time - midpt))).join(StatisticName). \
            filter(StatisticName.name == HEART_RATE,
                   StatisticName.owner == self.owner_in,
                   StatisticJournalInteger.time < finish,
                   StatisticJournalInteger.time >= start,
                   StatisticJournalInteger.value > 0).scalar()
        log.debug('M0: %s' % m0)
        if m0 and abs(m0 - QUARTER_DAY) < 0.25 * QUARTER_DAY:  # not evenly sampled
            all_hr = [row[0] for row in s.query(StatisticJournalInteger.value).join(StatisticName). \
                filter(StatisticName.name == HEART_RATE,
                       StatisticName.owner == self.owner_in,
                       StatisticJournalInteger.time < finish,
                       StatisticJournalInteger.time >= start,
                       StatisticJournalInteger.value > 0).all()]
            n = len(all_hr)
            rest_heart_rate = sorted(all_hr)[n // 10]  # 10th percentile
        else:
            log.info(f'Insufficient coverage for {REST_HR} for {start} - {finish}')
            rest_heart_rate = None
        daily_steps = s.query(func.sum(StatisticJournalInteger.value)).join(StatisticName). \
            filter(StatisticName.name == STEPS,
                   StatisticName.owner == self.owner_in,
                   StatisticJournalInteger.time < finish,
                   StatisticJournalInteger.time >= start).scalar()
        input_source_ids = [row[0] for row in s.query(MonitorJournal.id).
            filter(MonitorJournal.start <= finish,
                   MonitorJournal.finish >= start).all()]
        return input_source_ids, (rest_heart_rate, daily_steps)

    def _calculate_results(self, s, source, data, loader, start, finish):
        rest_heart_rate, daily_steps = data
        if rest_heart_rate:
            loader.add(REST_HR, BPM, summaries(AVG, CNT, MIN, MSR), None, source, rest_heart_rate,
                       start, StatisticJournalInteger)
        loader.add(DAILY_STEPS, STEPS_UNITS, summaries(SUM, AVG, CNT, MAX, MSR), None, source, daily_steps,
                   start, StatisticJournalInteger)




from math import pi, cos

from .tree import LinearMixin, BaseTree, QuadraticMixin, ExponentialMixin, CartesianMixin

RADIUS = 6371000
RADIAN = pi / 180


def norm180(x):
    while x > 180: x -= 360
    while x <= -180: x += 360
    return x


class LocalTangent:
    '''
    Assume a spherical earth and local linear approximations to convert from (lon, lat) to (x, y) in m.
    '''

    def __init__(self, point=None):
        self.__zero = None
        if point is not None:
            self.normalize(point)

    def normalize(self, point):
        if self.__zero is None:
            self.__zero = point
        zx, zy = self.__zero
        lon, lat = norm180(point[0] - zx), point[1] - zy
        return RADIUS * RADIAN * lon * cos(self.__zero[1]), RADIUS * RADIAN * lat

    def denormalize(self, point):
        zx, zy = self.__zero
        x, y = point
        return norm180(zx + x / (RADIUS * RADIAN * cos(self.__zero[1]))), zy + y / (RADIUS * RADIAN)


class SphericalMixin(CartesianMixin):

    def __init__(self, *args, **kargs):
        self.__plane = LocalTangent()
        super().__init__(*args, **kargs)

    def _normalize_point(self, point):
        return self.__plane.normalize(point)

    def _denormalize_point(self, point):
        return self.__plane.denormalize(point)


class SLRTree(LinearMixin, SphericalMixin, BaseTree): pass


class SQRTree(QuadraticMixin, SphericalMixin, BaseTree): pass


class SERTree(ExponentialMixin, SphericalMixin, BaseTree): pass


class Global:
    '''
    Tile a globe.
    '''

    def __init__(self, tree=SQRTree, n=36, favour_read=True):
        self.__tree = tree
        self.__n = n
        self.__trees = [[None] * n for _ in range(n)]
        self.__favour_read = favour_read

    def __norm_n(self, i):
        while i < 0: i += self.__n
        while i >= self.__n: i -= self.__n
        return i

    def __delegate(self, i, j):
        i, j = self.__norm_n(i), self.__norm_n(j)
        if self.__trees[i][j] is None:
            tree = self.__tree()
            # set the local tangent to the centre point
            bin_width = 360 / self.__n
            lon = (i + 0.5) * bin_width - 180
            lat = (j + 0.5) * bin_width - 180
            tree.add([(lon, lat)], None)
            tree.delete([(lon, lat)])
            self.__trees[i][j] = tree
        return self.__trees[i][j]

    def __delegates(self, points, read=True):
        lon, lat = points[0]
        i, j = int(self.__n * lon / 360), int(self.__n * lat / 180)
        for di in (-1, 0, 1):
            for dj in (-1, 0, 1):
                yield self.__delegate(i + di, j + dj)

    def get(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            yield from delegate.get(points, value=value, match=match, border=border)

    def get_items(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            yield from delegate.get_items(points, value=value, match=match, border=border)

    def add(self, points, value, border=None):
        for delegate in self.__delegates(points, read=False):
            delegate.add(points, value, border=border)

    def add_all(self, items, border=None):
        for (points, value) in items:
            for delegate in self.__delegates(points, read=False):
                delegate.add(points, value, border=border)

    def delete(self, points, value=None, match=None, border=None):
        '''
        This purposefully does not return number deleted.  If you are relying on that value,
        your code should not be using this wrapper.
        '''
        for delegate in self.__delegates(points):
            delegate.delete(points, value=value, match=match, border=border)

    def delete_one(self, points, value=None, match=None, border=None):
        for delegate in self.__delegates(points):
            delegate.delete_one(points, value=value, match=match, border=border)

    def __getitem__(self, points):
        return self.get(points)

    def __setitem__(self, points, value):
        self.add(points, value)

    def __delitem__(self, points):
        self.delete(points)

    def __bool__(self):
        return any(bool(t) for t in self.__trees)




from functools import lru_cache
from genericpath import exists
from math import floor
from os.path import join
from zipfile import ZipFile

import numpy as np

from ..squeal import Constant


SRTM1_DIR = 'SRTM1.Dir'
SAMPLES = 3601
# from view-source:http://dwtkns.com/srtm30m/
BASE_URL = 'http://e4ftl01.cr.usgs.gov/MEASURES/SRTMGL1.003/2000.02.11/'
EXTN = '.SRTMGL1.hgt.zip'


# lots of credit to https://github.com/aatishnn/srtm-python/blob/master/srtm.py
# (although that has bugs...)


@lru_cache(4)  # 4 means our tests are quick (and should tile a local patch)
def cached_file_reader(log, dir, flat, flon):
    if not exists(dir):
        raise Exception('SRTM1 directory %s missing' % dir)
    # https://wiki.openstreetmap.org/wiki/SRTM
    # The official 3-arc-second and 1-arc-second data for versions 2.1 and 3.0 are divided into 1°×1° data tiles.
    # The tiles are distributed as zip files containing HGT files labeled with the coordinate of the southwest cell.
    # For example, the file N20E100.hgt contains data from 20°N to 21°N and from 100°E to 101°E inclusive.
    root = '%s%02d%s%03d' % ('S' if flat < 0 else 'N', abs(flat), 'W' if flon < 0 else 'E', abs(flon))
    hgt_file = root + '.hgt'
    hgt_path = join(dir, hgt_file)
    zip_path = join(dir, root + EXTN)
    if exists(hgt_path):
        log.debug('Reading %s' % hgt_path)
        with open(hgt_path, 'rb') as input:
            data = input.read()
    elif exists(zip_path):
        log.debug('Reading %s' % zip_path)
        with open(zip_path, 'rb') as input:
            zip = ZipFile(input)
            log.debug('Found %s' % zip.filelist)
            data = zip.open(hgt_file).read()
    else:
        # i tried automating download, but couldn't get ouath2 to work
        log.warning('Download %s' % BASE_URL + root + EXTN)
        raise Exception('Missing %s' % hgt_file)
    return np.flip(np.frombuffer(data, np.dtype('>i2'), SAMPLES * SAMPLES).reshape((SAMPLES, SAMPLES)), 0)


class ElevationSupport:

    def __init__(self, log, dir, reader=cached_file_reader):
        self._log = log
        self._dir = dir
        self._reader = reader

    def _lookup(self, lat, lon):
        flat, flon = floor(lat), floor(lon)
        # construct the path in the reader so it's skipped if we hit the cache
        return flat, flon, self._reader(self._log, self._dir, flat, flon)


def elevation_from_constant(log, s, interp, dir_name=SRTM1_DIR):
    try:
        dir = Constant.get(s, dir_name).at(s).value
    except:
        log.warning('STRM1 config - define %s in constants for elevation data' % dir_name)
        dir = None
    return interp(log, dir)




import datetime as dt
from collections import namedtuple

import pandas as pd

from .lib import inplace_decay
from ..lib.date import round_hour

IMPULSE_3600 = 'Impulse / 3600s'
DecayModel = namedtuple('DecayModel', 'start, zero, scale, period, input, output')


def pre_calc(source, model, start=None, finish=None, target=None):
    # sum into 1-hour blocks (the decay is much longer period so this has little effect on accuracy
    # but saves time) and then add target indices
    # target here is intended for when fitting to data - that's the target data being fitted
    # (so we generate data at the right times)
    if target is not None:
        if start:
            start = min(start, target.index[0])
        else:
            start = target.index[0]
        if finish:
            finish = max(finish, target.index[-1])
        else:
            finish = target.index[-1]
    start, finish = round_hour(start, up=False), round_hour(finish, up=True)
    data = source.resample('1h', label='right').sum()
    data = data.loc[:, [model.input]]
    data.rename(columns={model.input: IMPULSE_3600}, inplace=True)
    times = pd.DataFrame(index=pd.date_range(start=start, end=finish, freq='1H'), columns=[model.output])
    data = data.join(times, how='outer', sort=True)
    t0 = data.index[0]
    initial = pd.DataFrame({IMPULSE_3600: model.start, model.output: 0}, index=[t0 - dt.timedelta(hours=1)])
    data = initial.append(data, sort=True)
    data.fillna(0, inplace=True)
    return data


def calc(data, model):
    data[model.output] = model.zero + data[IMPULSE_3600] * model.scale
    inplace_decay(data, model.output, model.period)
    return data
